#!/usr/bin/env python3

from .build import build_etree
from .forms import parse_form
from .html import extract_js_str, make_all_urls_absolute, parse_html_etree
from .text import extract_multiline_text, extract_text, normalize_spaces
from .utils import Element, detach, is_element, tostring
from .xml import parse_xml_etree



#!/usr/bin/env python3

# standards
import re
from typing import Any, Dict, Tuple, no_type_check

# 3rd parties
import lxml.etree as ET

# klon
from .utils import is_element


def build_etree(tag: str, *args) -> ET._Element:
    if not isinstance(tag, str):
        raise ValueError(f'Tag must be a str, got {tag!r}')
    attrib, args = _compile_attrib(*args)
    tag, attrib = _parse_css_style_tags(tag, attrib)
    element = ET.Element(tag, attrib)
    _append_children(element, args)
    return element


def _compile_attrib(*args) -> Tuple[Dict, Tuple[Any, ...]]:
    attrib = {}
    if len(args) > 0 and isinstance(args[0], dict):
        attrib.update(args[0])
        args = args[1:]
    return attrib, args


def _parse_css_style_tags(tag: str, attrib: Dict) -> Tuple[str, Dict]:
    match = re.search(r'^(.+?)(\#|\.)(.+)', tag)
    if match:
        tag, key, value = match.groups()
        key = {'#': 'id', '.': 'class'}[key]
        attrib[key] = value
    return tag, attrib


@no_type_check  # until lxml-stubs improves
def _append_children(element: ET._Element, args: Tuple[Any, ...]) -> None:
    text_anchor = None
    for child in args:
        if child in (None, (), []):
            pass
        elif isinstance(child, str):
            if text_anchor is None:
                element.text = (element.text or '') + child
            else:
                text_anchor.tail = (text_anchor.tail or '') + child
        else:
            if not is_element(child):
                child = build_etree(*child)
            element.append(child)
            text_anchor = child



#!/usr/bin/env python3

# standards
from typing import Any, Type, Union, no_type_check, overload

# 3rd parties
import lxml.etree as ET


Element = Union[ET._Element]  # this is exported, and can be used for type annotations


@no_type_check  # until lxml-stubs improves
def detach(node: Element, *, reattach_tail: bool = True) -> Element:
    parent = node.getparent()
    if parent is None:
        raise Exception('Node has no parent')
    if reattach_tail and node.tail:
        prev = node.getprevious()
        if prev is not None:
            prev.text = (prev.text + node.tail) if prev.text else node.tail
        else:
            parent.text = (parent.text + node.tail) if parent.text else node.tail
    parent.remove(node)
    return node


def is_element(obj: Any) -> bool:
    return isinstance(obj, ET._Element)


@overload
def tostring(etree: Union[Element, ET._ElementUnicodeResult], encoding: Type[str] = str, **kwargs) -> str:
    ...
@overload
def tostring(etree: Union[Element, ET._ElementUnicodeResult], encoding: str, **kwargs) -> bytes:
    ...
def tostring(etree, encoding=str, **kwargs):
    if isinstance(etree, ET._ElementUnicodeResult):
        text = str(etree)
        if encoding is not str:
            return text.encode(encoding)
        return text
    return ET.tostring(etree, encoding=encoding, **kwargs).strip()



#!/usr/bin/env python3

# standards
import re
from typing import List, no_type_check

# 3rd parties
import lxml.etree as ET


BLOCK_TAGS = frozenset([
    # This list taken from https://developer.mozilla.org/en-US/docs/Web/HTML/Block-level_elements
    'address', 'article', 'aside', 'blockquote', 'details', 'dialog', 'dd', 'div', 'dl', 'dt', 'fieldset', 'figcaption', 'figure',
    'figcaption', 'footer', 'form', 'h1', 'h2', 'h3', 'h4', 'h5', 'h6', 'header', 'hgroup', 'hr', 'li', 'main', 'nav', 'ol', 'p',
    'pre', 'section', 'table', 'ul'
])


NON_CONTENT_TAGS = frozenset([
    'head', 'script', 'style', 'svg',
])


def extract_text(etree: ET._Element, *, multiline: bool = False) -> str:
    if isinstance(etree, ET._ElementUnicodeResult):
        return normalize_spaces(etree)

    # using a list rather than making _walk() a yielding generator makes it about 5% faster
    parts: List[str] = []
    _walk(etree, parts)
    text = ''.join(parts)

    if multiline:
        return re.sub(
            r'\s+',
            lambda m: '\n\n' if '\n\n' in m.group() else '\n' if '\n' in m.group() else ' ',
            text
        ).strip()
    else:
        return normalize_spaces(text)


@no_type_check  # until lxml-stubs improves
def _walk(node: ET._Element, parts: List[str]) -> None:
    if node.tag in NON_CONTENT_TAGS or isinstance(node, ET._Comment):
        return

    if node.tag == 'br':
        parts.append('\n')
    elif node.tag in BLOCK_TAGS:
        parts.append('\n\n')

    if node.text:
        parts.append(re.sub(r'\s+', ' ', node.text))
    for child in node:
        _walk(child, parts)

        if child.tag in BLOCK_TAGS:
            parts.append('\n\n')
        if child.tail:
            parts.append(re.sub(r'\s+', ' ', child.tail))


def extract_multiline_text(etree: ET._Element) -> str:
    return extract_text(etree, multiline=True)


def normalize_spaces(text: str) -> str:
    return re.sub(r'\s+', ' ', text).strip()



#!/usr/bin/env python3

# standards
from pathlib import Path
import re
import setuptools


README_FILE = Path(__file__).parent / 'README.md'
LONG_DESCRIPTION = README_FILE.read_text('UTF-8')
LONG_DESCRIPTION = re.sub(
    r'(?<=\]\()(?!http)',
    'https://github.com/saintamh/klon/tree/master/',
    LONG_DESCRIPTION,
)

_version_match = re.search(
    r"KLON_VERSION = \'(.+)\'",
    (Path(__file__).parent / 'klon' / 'version.py').read_text('UTF-8'),
)
if not _version_match:
    raise ValueError('Version not found')
KLON_VERSION = _version_match.group(1)


setuptools.setup(
    name='klon',
    version=KLON_VERSION,
    author='Hervé Saint-Amand',
    author_email='klon@saintamh.org',
    description='Utilities for building and manipulating ElementTrees',
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    url='https://github.com/saintamh/klon/',
    packages=setuptools.find_packages(),
    package_data={'klon': ['py.typed']},
    install_requires=[
        'lxml>=4,<5',
        'lxml-stubs>=0.1,<0.2',  # so that users can type-check their calls to Klon
        'requests>=2.25,<3',
    ],
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Topic :: Text Processing :: Markup :: HTML',
        'Topic :: Text Processing :: Markup :: XML',
    ],
    zip_safe=False,  # https://mypy.readthedocs.io/en/latest/installed_packages.html#creating-pep-561-compatible-packages
)



#!/usr/bin/env python3

# 3rd parties
import pytest

# klon
from klon import build_etree, make_all_urls_absolute, tostring


@pytest.mark.parametrize(
    'base_url, etree, expected',
    [
        (
            # relative URLs are made absolute
            'http://example.com/',
            build_etree('div', ['a', {'href': 'relative/link'}]),
            build_etree('div', ['a', {'href': 'http://example.com/relative/link'}]),
        ),
        (
            # absolute URLs are untouched
            'http://example.com/',
            build_etree('div', ['a', {'href': 'http://absolute/link'}]),
            build_etree('div', ['a', {'href': 'http://absolute/link'}]),
        ),
        (
            # an https:// URL doesn't override an http:// URL
            'https://example.com/',
            build_etree('div', ['a', {'href': 'http://absolute/link'}]),
            build_etree('div', ['a', {'href': 'http://absolute/link'}]),
        ),
        (
            # an #anchor link gets merged too
            'http://example.com/',
            build_etree('div', ['a', {'href': '#anchor'}]),
            build_etree('div', ['a', {'href': 'http://example.com/#anchor'}]),
        ),
        (
            # not just ahrefs are updated -- see full list in klon/html.py
            'http://example.com/',
            build_etree('div', ['img', {'src': 'image.jpg'}]),
            build_etree('div', ['img', {'src': 'http://example.com/image.jpg'}]),
        ),
        (
            # the root element is updated too
            'http://example.com/',
            build_etree('a', {'href': 'relative/link'}),
            build_etree('a', {'href': 'http://example.com/relative/link'}),
        ),
        (
            # if the URL attribute is missing we don't complain
            'http://example.com/',
            build_etree('div', ['a', {'name': 'rover'}]),
            build_etree('div', ['a', {'name': 'rover'}]),
        ),
        (
            # empty links are replaced by the full base URL, as a browser does
            'http://example.com/',
            build_etree('div', ['a', {'href': ''}]),
            build_etree('div', ['a', {'href': 'http://example.com/'}]),
        ),
    ]
)
def test_make_all_urls_absolute(base_url, etree, expected):
    make_all_urls_absolute(base_url, etree)
    assert tostring(etree) == tostring(expected)
